// -*- C++ -*-
//
// SimpleColourBasis2.h is a part of Herwig - A multi-purpose Monte Carlo event generator
// Copyright (C) 2002-2019 The Herwig Collaboration
//
// Herwig is licenced under version 3 of the GPL, see COPYING for details.
// Please respect the MCnet academic guidelines, see GUIDELINES for details.
//
//
// This is the implementation of the non-inlined, non-templated member
// functions of the SimpleColourBasis2 class.
//

#include "SimpleColourBasis2.h"
#include "ThePEG/Interface/ClassDocumentation.h"
#include "ThePEG/EventRecord/Particle.h"
#include "ThePEG/Repository/UseRandom.h"
#include "ThePEG/Repository/EventGenerator.h"
#include "ThePEG/Utilities/DescribeClass.h"
#include "ThePEG/StandardModel/StandardModelBase.h"
#include "ThePEG/Persistency/PersistentOStream.h"
#include "ThePEG/Persistency/PersistentIStream.h"

using namespace Herwig;

IBPtr SimpleColourBasis2::clone() const {
  return new_ptr(*this);
}

IBPtr SimpleColourBasis2::fullclone() const {
  return new_ptr(*this);
}

size_t SimpleColourBasis2::prepareBasis(const vector<PDT::Colour>& basis) {

  if ( id33bar.empty() )
    makeIds();

  if ( basis == id88 ) {
    return 1;
  }

  if ( basis == id33bar ) {
    return 1;
  }

  if ( basis == id888 ) {
    return 2;
  }

  if ( basis == id33bar8 ) {
    return 1;
  }

  if ( basis == id8888 ) {
    return 9;
  }

  if ( basis == id33bar88 ) {
    return 3;
  }

  if ( basis == id33bar33bar ) {
    return 2;
  }

  if ( basis == id88888 ) {
    return 44;
  }

  if ( basis == id33bar888 ) {
    return 11;
  }

  if ( basis == id33bar33bar8 ) {
    return 4;
  }

  throw Exception() << "SimpleColourBasis2::prepareBasis(): Cannot handle colour configuration" << Exception::runerror;

  return 0;

}

double SimpleColourBasis2::scalarProduct(size_t a, size_t b,
					const vector<PDT::Colour>& basis) const {

  if ( id33bar.empty() )
    makeIds();

  double Nc = SM().Nc();
  double Nc2 = sqr(Nc);
  double Nc3 = Nc*Nc2;
  double Nc4 = sqr(Nc2);
  double Nc5 = Nc*Nc4;
  double Nc6 = Nc2*Nc4;
  double Nc8 = Nc2*Nc6;

  if ( a > b )
    swap(a,b);

  if ( !largeN() ) {

    if ( basis == id88 ) {
      if( a == 0 && b == 0  )
	return (-1 + Nc2)/4.;
    }

    if ( basis == id33bar ) {
      if( a == 0 && b == 0  )
	return Nc;
    }

    if ( basis == id888 ) {
      if( ( a == 0 && b == 0 ) ||
	  ( a == 1 && b == 1 )  )
	return (2 - 3*Nc2 + Nc4)/(8.*Nc);
      if( ( a == 0 ) && ( b == 1 )  )
	return -(-1 + Nc2)/(4.*Nc);
    }

    if ( basis == id33bar8 ) {
      if( a == 0 && b == 0  )
	return (-1 + Nc2)/2.;
    }

    if ( basis == id8888 ) {
      if( ( a == 0 && b == 0 ) ||
	  ( a == 1 && b == 1 ) ||
	  ( a == 2 && b == 2 )  )
	return sqr(-1 + Nc2)/16.;
      if( ( a == 0 && b == 1 ) ||
	  ( a == 0 && b == 2 ) ||
	  ( a == 1 && b == 2 )  )
	return (-1 + Nc2)/16.;
      if( ( a == 0 && b == 3 ) ||
	  ( a == 0 && b == 5 ) ||
	  ( a == 0 && b == 7 ) ||
	  ( a == 0 && b == 8 ) ||
	  ( a == 1 && b == 4 ) ||
	  ( a == 1 && b == 5 ) ||
	  ( a == 1 && b == 6 ) ||
	  ( a == 1 && b == 7 ) ||
	  ( a == 2 && b == 3 ) ||
	  ( a == 2 && b == 4 ) ||
	  ( a == 2 && b == 6 ) ||
	  ( a == 2 && b == 8 )  )
	return sqr(-1 + Nc2)/(16.*Nc);
      if( ( a == 0 && b == 4 ) ||
	  ( a == 0 && b == 6 ) ||
	  ( a == 1 && b == 3 ) ||
	  ( a == 1 && b == 8 ) ||
	  ( a == 2 && b == 5 ) ||
	  ( a == 2 && b == 7 )  )
	return -(-1 + Nc2)/(16.*Nc);
      if( ( a == 3 && b == 3 ) ||
	  ( a == 4 && b == 4 ) ||
	  ( a == 5 && b == 5 ) ||
	  ( a == 6 && b == 6 ) ||
	  ( a == 7 && b == 7 ) ||
	  ( a == 8 && b == 8 )  )
	return (-3 + 6*Nc2 - 4*Nc4 + Nc6)/(16.*Nc2);
      if( ( a == 3 && b == 4 ) ||
	  ( a == 3 && b == 5 ) ||
	  ( a == 3 && b == 6 ) ||
	  ( a == 3 && b == 7 ) ||
	  ( a == 4 && b == 5 ) ||
	  ( a == 4 && b == 7 ) ||
	  ( a == 4 && b == 8 ) ||
	  ( a == 5 && b == 6 ) ||
	  ( a == 5 && b == 8 ) ||
	  ( a == 6 && b == 7 ) ||
	  ( a == 6 && b == 8 ) ||
	  ( a == 7 && b == 8 )  )
	return (4 - 3/Nc2 - Nc2)/16.;
      if( ( a == 3 && b == 8 ) ||
	  ( a == 4 && b == 6 ) ||
	  ( a == 5 && b == 7 )  )
	return (-3 + 2*Nc2 + Nc4)/(16.*Nc2);
    }

    if ( basis == id33bar88 ) {
      if( a == 0 && b == 0  )
	return (Nc*(-1 + Nc2))/4.;
      if( ( a == 0 && b == 1 ) ||
	  ( a == 0 && b == 2 )  )
	return (-1 + Nc2)/4.;
      if( ( a == 1 && b == 1 ) ||
	  ( a == 2 && b == 2 )  )
	return sqr(-1 + Nc2)/(4.*Nc);
      if( a == 1 && b == 2  )
	return -(-1 + Nc2)/(4.*Nc);
    }

    if ( basis == id33bar33bar ) {
      if( ( a == 0 && b == 0 ) ||
	  ( a == 1 && b == 1 )  )
	return Nc2;
      if( a == 0 && b == 1  )
	return Nc;
    }

    if ( basis == id88888 ) {
      if( ( a == 0 && b == 0 ) ||
	  ( a == 1 && b == 1 ) ||
	  ( a == 2 && b == 2 ) ||
	  ( a == 3 && b == 3 ) ||
	  ( a == 4 && b == 4 ) ||
	  ( a == 5 && b == 5 ) ||
	  ( a == 6 && b == 6 ) ||
	  ( a == 7 && b == 7 ) ||
	  ( a == 8 && b == 8 ) ||
	  ( a == 9 && b == 9 ) ||
	  ( a == 10 && b == 10 ) ||
	  ( a == 11 && b == 11 ) ||
	  ( a == 12 && b == 12 ) ||
	  ( a == 13 && b == 13 ) ||
	  ( a == 14 && b == 14 ) ||
	  ( a == 15 && b == 15 ) ||
	  ( a == 16 && b == 16 ) ||
	  ( a == 17 && b == 17 ) ||
	  ( a == 18 && b == 18 ) ||
	  ( a == 19 && b == 19 )  )
	return ((-2 + Nc2)*sqr(-1 + Nc2))/(32.*Nc);
      if( ( a == 0 && b == 1 ) ||
	  ( a == 0 && b == 2 ) ||
	  ( a == 0 && b == 7 ) ||
	  ( a == 0 && b == 10 ) ||
	  ( a == 0 && b == 12 ) ||
	  ( a == 0 && b == 13 ) ||
	  ( a == 1 && b == 2 ) ||
	  ( a == 1 && b == 4 ) ||
	  ( a == 1 && b == 11 ) ||
	  ( a == 1 && b == 14 ) ||
	  ( a == 1 && b == 15 ) ||
	  ( a == 2 && b == 5 ) ||
	  ( a == 2 && b == 8 ) ||
	  ( a == 2 && b == 16 ) ||
	  ( a == 2 && b == 17 ) ||
	  ( a == 3 && b == 4 ) ||
	  ( a == 3 && b == 5 ) ||
	  ( a == 3 && b == 6 ) ||
	  ( a == 3 && b == 9 ) ||
	  ( a == 3 && b == 14 ) ||
	  ( a == 3 && b == 16 ) ||
	  ( a == 4 && b == 5 ) ||
	  ( a == 4 && b == 11 ) ||
	  ( a == 4 && b == 12 ) ||
	  ( a == 4 && b == 18 ) ||
	  ( a == 5 && b == 8 ) ||
	  ( a == 5 && b == 13 ) ||
	  ( a == 5 && b == 19 ) ||
	  ( a == 6 && b == 7 ) ||
	  ( a == 6 && b == 8 ) ||
	  ( a == 6 && b == 9 ) ||
	  ( a == 6 && b == 12 ) ||
	  ( a == 6 && b == 17 ) ||
	  ( a == 7 && b == 8 ) ||
	  ( a == 7 && b == 10 ) ||
	  ( a == 7 && b == 14 ) ||
	  ( a == 7 && b == 19 ) ||
	  ( a == 8 && b == 15 ) ||
	  ( a == 8 && b == 18 ) ||
	  ( a == 9 && b == 10 ) ||
	  ( a == 9 && b == 11 ) ||
	  ( a == 9 && b == 13 ) ||
	  ( a == 9 && b == 15 ) ||
	  ( a == 10 && b == 11 ) ||
	  ( a == 10 && b == 16 ) ||
	  ( a == 10 && b == 18 ) ||
	  ( a == 11 && b == 17 ) ||
	  ( a == 11 && b == 19 ) ||
	  ( a == 12 && b == 13 ) ||
	  ( a == 12 && b == 17 ) ||
	  ( a == 12 && b == 18 ) ||
	  ( a == 13 && b == 15 ) ||
	  ( a == 13 && b == 19 ) ||
	  ( a == 14 && b == 15 ) ||
	  ( a == 14 && b == 16 ) ||
	  ( a == 14 && b == 19 ) ||
	  ( a == 15 && b == 18 ) ||
	  ( a == 16 && b == 17 ) ||
	  ( a == 16 && b == 18 ) ||
	  ( a == 17 && b == 19 )  )
	return (2 - 3*Nc2 + Nc4)/(32.*Nc);
      if( ( a == 0 && b == 3 ) ||
	  ( a == 1 && b == 6 ) ||
	  ( a == 2 && b == 9 ) ||
	  ( a == 4 && b == 7 ) ||
	  ( a == 5 && b == 10 ) ||
	  ( a == 8 && b == 11 ) ||
	  ( a == 12 && b == 14 ) ||
	  ( a == 13 && b == 16 ) ||
	  ( a == 15 && b == 17 ) ||
	  ( a == 18 && b == 19 )  )
	return -sqr(-1 + Nc2)/(16.*Nc);
      if( ( a == 0 && b == 4 ) ||
	  ( a == 0 && b == 5 ) ||
	  ( a == 0 && b == 6 ) ||
	  ( a == 0 && b == 9 ) ||
	  ( a == 0 && b == 14 ) ||
	  ( a == 0 && b == 16 ) ||
	  ( a == 1 && b == 3 ) ||
	  ( a == 1 && b == 7 ) ||
	  ( a == 1 && b == 8 ) ||
	  ( a == 1 && b == 9 ) ||
	  ( a == 1 && b == 12 ) ||
	  ( a == 1 && b == 17 ) ||
	  ( a == 2 && b == 3 ) ||
	  ( a == 2 && b == 6 ) ||
	  ( a == 2 && b == 10 ) ||
	  ( a == 2 && b == 11 ) ||
	  ( a == 2 && b == 13 ) ||
	  ( a == 2 && b == 15 ) ||
	  ( a == 3 && b == 7 ) ||
	  ( a == 3 && b == 10 ) ||
	  ( a == 3 && b == 12 ) ||
	  ( a == 3 && b == 13 ) ||
	  ( a == 4 && b == 6 ) ||
	  ( a == 4 && b == 8 ) ||
	  ( a == 4 && b == 10 ) ||
	  ( a == 4 && b == 14 ) ||
	  ( a == 4 && b == 19 ) ||
	  ( a == 5 && b == 7 ) ||
	  ( a == 5 && b == 9 ) ||
	  ( a == 5 && b == 11 ) ||
	  ( a == 5 && b == 16 ) ||
	  ( a == 5 && b == 18 ) ||
	  ( a == 6 && b == 11 ) ||
	  ( a == 6 && b == 14 ) ||
	  ( a == 6 && b == 15 ) ||
	  ( a == 7 && b == 11 ) ||
	  ( a == 7 && b == 12 ) ||
	  ( a == 7 && b == 18 ) ||
	  ( a == 8 && b == 9 ) ||
	  ( a == 8 && b == 10 ) ||
	  ( a == 8 && b == 17 ) ||
	  ( a == 8 && b == 19 ) ||
	  ( a == 9 && b == 16 ) ||
	  ( a == 9 && b == 17 ) ||
	  ( a == 10 && b == 13 ) ||
	  ( a == 10 && b == 19 ) ||
	  ( a == 11 && b == 15 ) ||
	  ( a == 11 && b == 18 ) ||
	  ( a == 12 && b == 15 ) ||
	  ( a == 12 && b == 16 ) ||
	  ( a == 12 && b == 19 ) ||
	  ( a == 13 && b == 14 ) ||
	  ( a == 13 && b == 17 ) ||
	  ( a == 13 && b == 18 ) ||
	  ( a == 14 && b == 17 ) ||
	  ( a == 14 && b == 18 ) ||
	  ( a == 15 && b == 16 ) ||
	  ( a == 15 && b == 19 ) ||
	  ( a == 16 && b == 19 ) ||
	  ( a == 17 && b == 18 )  )
	return -(-1 + Nc2)/(16.*Nc);
      if( ( a == 0 && b == 8 ) ||
	  ( a == 0 && b == 11 ) ||
	  ( a == 0 && b == 15 ) ||
	  ( a == 0 && b == 17 ) ||
	  ( a == 0 && b == 18 ) ||
	  ( a == 0 && b == 19 ) ||
	  ( a == 1 && b == 5 ) ||
	  ( a == 1 && b == 10 ) ||
	  ( a == 1 && b == 13 ) ||
	  ( a == 1 && b == 16 ) ||
	  ( a == 1 && b == 18 ) ||
	  ( a == 1 && b == 19 ) ||
	  ( a == 2 && b == 4 ) ||
	  ( a == 2 && b == 7 ) ||
	  ( a == 2 && b == 12 ) ||
	  ( a == 2 && b == 14 ) ||
	  ( a == 2 && b == 18 ) ||
	  ( a == 2 && b == 19 ) ||
	  ( a == 3 && b == 8 ) ||
	  ( a == 3 && b == 11 ) ||
	  ( a == 3 && b == 15 ) ||
	  ( a == 3 && b == 17 ) ||
	  ( a == 3 && b == 18 ) ||
	  ( a == 3 && b == 19 ) ||
	  ( a == 4 && b == 9 ) ||
	  ( a == 4 && b == 13 ) ||
	  ( a == 4 && b == 15 ) ||
	  ( a == 4 && b == 16 ) ||
	  ( a == 4 && b == 17 ) ||
	  ( a == 5 && b == 6 ) ||
	  ( a == 5 && b == 12 ) ||
	  ( a == 5 && b == 14 ) ||
	  ( a == 5 && b == 15 ) ||
	  ( a == 5 && b == 17 ) ||
	  ( a == 6 && b == 10 ) ||
	  ( a == 6 && b == 13 ) ||
	  ( a == 6 && b == 16 ) ||
	  ( a == 6 && b == 18 ) ||
	  ( a == 6 && b == 19 ) ||
	  ( a == 7 && b == 9 ) ||
	  ( a == 7 && b == 13 ) ||
	  ( a == 7 && b == 15 ) ||
	  ( a == 7 && b == 16 ) ||
	  ( a == 7 && b == 17 ) ||
	  ( a == 8 && b == 12 ) ||
	  ( a == 8 && b == 13 ) ||
	  ( a == 8 && b == 14 ) ||
	  ( a == 8 && b == 16 ) ||
	  ( a == 9 && b == 12 ) ||
	  ( a == 9 && b == 14 ) ||
	  ( a == 9 && b == 18 ) ||
	  ( a == 9 && b == 19 ) ||
	  ( a == 10 && b == 12 ) ||
	  ( a == 10 && b == 14 ) ||
	  ( a == 10 && b == 15 ) ||
	  ( a == 10 && b == 17 ) ||
	  ( a == 11 && b == 12 ) ||
	  ( a == 11 && b == 13 ) ||
	  ( a == 11 && b == 14 ) ||
	  ( a == 11 && b == 16 )  )
	return 0;
      if( ( a == 0 && b == 20 ) ||
	  ( a == 0 && b == 21 ) ||
	  ( a == 0 && b == 23 ) ||
	  ( a == 0 && b == 25 ) ||
	  ( a == 0 && b == 36 ) ||
	  ( a == 0 && b == 42 ) ||
	  ( a == 1 && b == 21 ) ||
	  ( a == 1 && b == 22 ) ||
	  ( a == 1 && b == 23 ) ||
	  ( a == 1 && b == 24 ) ||
	  ( a == 1 && b == 30 ) ||
	  ( a == 1 && b == 40 ) ||
	  ( a == 2 && b == 20 ) ||
	  ( a == 2 && b == 22 ) ||
	  ( a == 2 && b == 24 ) ||
	  ( a == 2 && b == 25 ) ||
	  ( a == 2 && b == 28 ) ||
	  ( a == 2 && b == 34 ) ||
	  ( a == 3 && b == 26 ) ||
	  ( a == 3 && b == 27 ) ||
	  ( a == 3 && b == 29 ) ||
	  ( a == 3 && b == 31 ) ||
	  ( a == 3 && b == 37 ) ||
	  ( a == 3 && b == 43 ) ||
	  ( a == 4 && b == 24 ) ||
	  ( a == 4 && b == 27 ) ||
	  ( a == 4 && b == 28 ) ||
	  ( a == 4 && b == 29 ) ||
	  ( a == 4 && b == 30 ) ||
	  ( a == 4 && b == 38 ) ||
	  ( a == 5 && b == 22 ) ||
	  ( a == 5 && b == 26 ) ||
	  ( a == 5 && b == 28 ) ||
	  ( a == 5 && b == 30 ) ||
	  ( a == 5 && b == 31 ) ||
	  ( a == 5 && b == 32 ) ||
	  ( a == 6 && b == 31 ) ||
	  ( a == 6 && b == 32 ) ||
	  ( a == 6 && b == 33 ) ||
	  ( a == 6 && b == 35 ) ||
	  ( a == 6 && b == 37 ) ||
	  ( a == 6 && b == 41 ) ||
	  ( a == 7 && b == 25 ) ||
	  ( a == 7 && b == 33 ) ||
	  ( a == 7 && b == 34 ) ||
	  ( a == 7 && b == 35 ) ||
	  ( a == 7 && b == 36 ) ||
	  ( a == 7 && b == 39 ) ||
	  ( a == 8 && b == 20 ) ||
	  ( a == 8 && b == 26 ) ||
	  ( a == 8 && b == 32 ) ||
	  ( a == 8 && b == 34 ) ||
	  ( a == 8 && b == 36 ) ||
	  ( a == 8 && b == 37 ) ||
	  ( a == 9 && b == 29 ) ||
	  ( a == 9 && b == 35 ) ||
	  ( a == 9 && b == 38 ) ||
	  ( a == 9 && b == 39 ) ||
	  ( a == 9 && b == 41 ) ||
	  ( a == 9 && b == 43 ) ||
	  ( a == 10 && b == 23 ) ||
	  ( a == 10 && b == 33 ) ||
	  ( a == 10 && b == 39 ) ||
	  ( a == 10 && b == 40 ) ||
	  ( a == 10 && b == 41 ) ||
	  ( a == 10 && b == 42 ) ||
	  ( a == 11 && b == 21 ) ||
	  ( a == 11 && b == 27 ) ||
	  ( a == 11 && b == 38 ) ||
	  ( a == 11 && b == 40 ) ||
	  ( a == 11 && b == 42 ) ||
	  ( a == 11 && b == 43 ) ||
	  ( a == 12 && b == 20 ) ||
	  ( a == 12 && b == 28 ) ||
	  ( a == 12 && b == 32 ) ||
	  ( a == 12 && b == 38 ) ||
	  ( a == 12 && b == 41 ) ||
	  ( a == 12 && b == 42 ) ||
	  ( a == 13 && b == 21 ) ||
	  ( a == 13 && b == 30 ) ||
	  ( a == 13 && b == 32 ) ||
	  ( a == 13 && b == 35 ) ||
	  ( a == 13 && b == 36 ) ||
	  ( a == 13 && b == 38 ) ||
	  ( a == 14 && b == 22 ) ||
	  ( a == 14 && b == 26 ) ||
	  ( a == 14 && b == 34 ) ||
	  ( a == 14 && b == 39 ) ||
	  ( a == 14 && b == 40 ) ||
	  ( a == 14 && b == 43 ) ||
	  ( a == 15 && b == 23 ) ||
	  ( a == 15 && b == 26 ) ||
	  ( a == 15 && b == 29 ) ||
	  ( a == 15 && b == 30 ) ||
	  ( a == 15 && b == 36 ) ||
	  ( a == 15 && b == 39 ) ||
	  ( a == 16 && b == 24 ) ||
	  ( a == 16 && b == 27 ) ||
	  ( a == 16 && b == 33 ) ||
	  ( a == 16 && b == 34 ) ||
	  ( a == 16 && b == 37 ) ||
	  ( a == 16 && b == 40 ) ||
	  ( a == 17 && b == 25 ) ||
	  ( a == 17 && b == 27 ) ||
	  ( a == 17 && b == 28 ) ||
	  ( a == 17 && b == 31 ) ||
	  ( a == 17 && b == 33 ) ||
	  ( a == 17 && b == 42 ) ||
	  ( a == 18 && b == 20 ) ||
	  ( a == 18 && b == 23 ) ||
	  ( a == 18 && b == 24 ) ||
	  ( a == 18 && b == 29 ) ||
	  ( a == 18 && b == 37 ) ||
	  ( a == 18 && b == 41 ) ||
	  ( a == 19 && b == 21 ) ||
	  ( a == 19 && b == 22 ) ||
	  ( a == 19 && b == 25 ) ||
	  ( a == 19 && b == 31 ) ||
	  ( a == 19 && b == 35 ) ||
	  ( a == 19 && b == 43 )  )
	return ((-2 + Nc2)*sqr(-1 + Nc2))/(32.*Nc2);
      if( ( a == 0 && b == 22 ) ||
	  ( a == 0 && b == 24 ) ||
	  ( a == 0 && b == 32 ) ||
	  ( a == 0 && b == 33 ) ||
	  ( a == 0 && b == 38 ) ||
	  ( a == 0 && b == 39 ) ||
	  ( a == 1 && b == 20 ) ||
	  ( a == 1 && b == 25 ) ||
	  ( a == 1 && b == 26 ) ||
	  ( a == 1 && b == 27 ) ||
	  ( a == 1 && b == 38 ) ||
	  ( a == 1 && b == 39 ) ||
	  ( a == 2 && b == 21 ) ||
	  ( a == 2 && b == 23 ) ||
	  ( a == 2 && b == 26 ) ||
	  ( a == 2 && b == 27 ) ||
	  ( a == 2 && b == 32 ) ||
	  ( a == 2 && b == 33 ) ||
	  ( a == 3 && b == 28 ) ||
	  ( a == 3 && b == 30 ) ||
	  ( a == 3 && b == 34 ) ||
	  ( a == 3 && b == 35 ) ||
	  ( a == 3 && b == 40 ) ||
	  ( a == 3 && b == 41 ) ||
	  ( a == 4 && b == 20 ) ||
	  ( a == 4 && b == 21 ) ||
	  ( a == 4 && b == 26 ) ||
	  ( a == 4 && b == 31 ) ||
	  ( a == 4 && b == 40 ) ||
	  ( a == 4 && b == 41 ) ||
	  ( a == 5 && b == 20 ) ||
	  ( a == 5 && b == 21 ) ||
	  ( a == 5 && b == 27 ) ||
	  ( a == 5 && b == 29 ) ||
	  ( a == 5 && b == 34 ) ||
	  ( a == 5 && b == 35 ) ||
	  ( a == 6 && b == 28 ) ||
	  ( a == 6 && b == 29 ) ||
	  ( a == 6 && b == 34 ) ||
	  ( a == 6 && b == 36 ) ||
	  ( a == 6 && b == 42 ) ||
	  ( a == 6 && b == 43 ) ||
	  ( a == 7 && b == 22 ) ||
	  ( a == 7 && b == 23 ) ||
	  ( a == 7 && b == 32 ) ||
	  ( a == 7 && b == 37 ) ||
	  ( a == 7 && b == 42 ) ||
	  ( a == 7 && b == 43 ) ||
	  ( a == 8 && b == 22 ) ||
	  ( a == 8 && b == 23 ) ||
	  ( a == 8 && b == 28 ) ||
	  ( a == 8 && b == 29 ) ||
	  ( a == 8 && b == 33 ) ||
	  ( a == 8 && b == 35 ) ||
	  ( a == 9 && b == 30 ) ||
	  ( a == 9 && b == 31 ) ||
	  ( a == 9 && b == 36 ) ||
	  ( a == 9 && b == 37 ) ||
	  ( a == 9 && b == 40 ) ||
	  ( a == 9 && b == 42 ) ||
	  ( a == 10 && b == 24 ) ||
	  ( a == 10 && b == 25 ) ||
	  ( a == 10 && b == 36 ) ||
	  ( a == 10 && b == 37 ) ||
	  ( a == 10 && b == 38 ) ||
	  ( a == 10 && b == 43 ) ||
	  ( a == 11 && b == 24 ) ||
	  ( a == 11 && b == 25 ) ||
	  ( a == 11 && b == 30 ) ||
	  ( a == 11 && b == 31 ) ||
	  ( a == 11 && b == 39 ) ||
	  ( a == 11 && b == 41 ) ||
	  ( a == 12 && b == 21 ) ||
	  ( a == 12 && b == 24 ) ||
	  ( a == 12 && b == 29 ) ||
	  ( a == 12 && b == 31 ) ||
	  ( a == 12 && b == 33 ) ||
	  ( a == 12 && b == 36 ) ||
	  ( a == 13 && b == 20 ) ||
	  ( a == 13 && b == 22 ) ||
	  ( a == 13 && b == 29 ) ||
	  ( a == 13 && b == 31 ) ||
	  ( a == 13 && b == 39 ) ||
	  ( a == 13 && b == 42 ) ||
	  ( a == 14 && b == 23 ) ||
	  ( a == 14 && b == 25 ) ||
	  ( a == 14 && b == 27 ) ||
	  ( a == 14 && b == 30 ) ||
	  ( a == 14 && b == 35 ) ||
	  ( a == 14 && b == 37 ) ||
	  ( a == 15 && b == 20 ) ||
	  ( a == 15 && b == 22 ) ||
	  ( a == 15 && b == 35 ) ||
	  ( a == 15 && b == 37 ) ||
	  ( a == 15 && b == 38 ) ||
	  ( a == 15 && b == 40 ) ||
	  ( a == 16 && b == 23 ) ||
	  ( a == 16 && b == 25 ) ||
	  ( a == 16 && b == 26 ) ||
	  ( a == 16 && b == 28 ) ||
	  ( a == 16 && b == 41 ) ||
	  ( a == 16 && b == 43 ) ||
	  ( a == 17 && b == 21 ) ||
	  ( a == 17 && b == 24 ) ||
	  ( a == 17 && b == 32 ) ||
	  ( a == 17 && b == 34 ) ||
	  ( a == 17 && b == 41 ) ||
	  ( a == 17 && b == 43 ) ||
	  ( a == 18 && b == 26 ) ||
	  ( a == 18 && b == 28 ) ||
	  ( a == 18 && b == 33 ) ||
	  ( a == 18 && b == 36 ) ||
	  ( a == 18 && b == 38 ) ||
	  ( a == 18 && b == 40 ) ||
	  ( a == 19 && b == 27 ) ||
	  ( a == 19 && b == 30 ) ||
	  ( a == 19 && b == 32 ) ||
	  ( a == 19 && b == 34 ) ||
	  ( a == 19 && b == 39 ) ||
	  ( a == 19 && b == 42 )  )
	return -(2 - 3*Nc2 + Nc4)/(32.*Nc2);
      if( ( a == 0 && b == 26 ) ||
	  ( a == 0 && b == 27 ) ||
	  ( a == 0 && b == 29 ) ||
	  ( a == 0 && b == 31 ) ||
	  ( a == 0 && b == 37 ) ||
	  ( a == 0 && b == 43 ) ||
	  ( a == 1 && b == 31 ) ||
	  ( a == 1 && b == 32 ) ||
	  ( a == 1 && b == 33 ) ||
	  ( a == 1 && b == 35 ) ||
	  ( a == 1 && b == 37 ) ||
	  ( a == 1 && b == 41 ) ||
	  ( a == 2 && b == 29 ) ||
	  ( a == 2 && b == 35 ) ||
	  ( a == 2 && b == 38 ) ||
	  ( a == 2 && b == 39 ) ||
	  ( a == 2 && b == 41 ) ||
	  ( a == 2 && b == 43 ) ||
	  ( a == 3 && b == 20 ) ||
	  ( a == 3 && b == 21 ) ||
	  ( a == 3 && b == 23 ) ||
	  ( a == 3 && b == 25 ) ||
	  ( a == 3 && b == 36 ) ||
	  ( a == 3 && b == 42 ) ||
	  ( a == 4 && b == 25 ) ||
	  ( a == 4 && b == 33 ) ||
	  ( a == 4 && b == 34 ) ||
	  ( a == 4 && b == 35 ) ||
	  ( a == 4 && b == 36 ) ||
	  ( a == 4 && b == 39 ) ||
	  ( a == 5 && b == 23 ) ||
	  ( a == 5 && b == 33 ) ||
	  ( a == 5 && b == 39 ) ||
	  ( a == 5 && b == 40 ) ||
	  ( a == 5 && b == 41 ) ||
	  ( a == 5 && b == 42 ) ||
	  ( a == 6 && b == 21 ) ||
	  ( a == 6 && b == 22 ) ||
	  ( a == 6 && b == 23 ) ||
	  ( a == 6 && b == 24 ) ||
	  ( a == 6 && b == 30 ) ||
	  ( a == 6 && b == 40 ) ||
	  ( a == 7 && b == 24 ) ||
	  ( a == 7 && b == 27 ) ||
	  ( a == 7 && b == 28 ) ||
	  ( a == 7 && b == 29 ) ||
	  ( a == 7 && b == 30 ) ||
	  ( a == 7 && b == 38 ) ||
	  ( a == 8 && b == 21 ) ||
	  ( a == 8 && b == 27 ) ||
	  ( a == 8 && b == 38 ) ||
	  ( a == 8 && b == 40 ) ||
	  ( a == 8 && b == 42 ) ||
	  ( a == 8 && b == 43 ) ||
	  ( a == 9 && b == 20 ) ||
	  ( a == 9 && b == 22 ) ||
	  ( a == 9 && b == 24 ) ||
	  ( a == 9 && b == 25 ) ||
	  ( a == 9 && b == 28 ) ||
	  ( a == 9 && b == 34 ) ||
	  ( a == 10 && b == 22 ) ||
	  ( a == 10 && b == 26 ) ||
	  ( a == 10 && b == 28 ) ||
	  ( a == 10 && b == 30 ) ||
	  ( a == 10 && b == 31 ) ||
	  ( a == 10 && b == 32 ) ||
	  ( a == 11 && b == 20 ) ||
	  ( a == 11 && b == 26 ) ||
	  ( a == 11 && b == 32 ) ||
	  ( a == 11 && b == 34 ) ||
	  ( a == 11 && b == 36 ) ||
	  ( a == 11 && b == 37 ) ||
	  ( a == 12 && b == 22 ) ||
	  ( a == 12 && b == 26 ) ||
	  ( a == 12 && b == 34 ) ||
	  ( a == 12 && b == 39 ) ||
	  ( a == 12 && b == 40 ) ||
	  ( a == 12 && b == 43 ) ||
	  ( a == 13 && b == 24 ) ||
	  ( a == 13 && b == 27 ) ||
	  ( a == 13 && b == 33 ) ||
	  ( a == 13 && b == 34 ) ||
	  ( a == 13 && b == 37 ) ||
	  ( a == 13 && b == 40 ) ||
	  ( a == 14 && b == 20 ) ||
	  ( a == 14 && b == 28 ) ||
	  ( a == 14 && b == 32 ) ||
	  ( a == 14 && b == 38 ) ||
	  ( a == 14 && b == 41 ) ||
	  ( a == 14 && b == 42 ) ||
	  ( a == 15 && b == 25 ) ||
	  ( a == 15 && b == 27 ) ||
	  ( a == 15 && b == 28 ) ||
	  ( a == 15 && b == 31 ) ||
	  ( a == 15 && b == 33 ) ||
	  ( a == 15 && b == 42 ) ||
	  ( a == 16 && b == 21 ) ||
	  ( a == 16 && b == 30 ) ||
	  ( a == 16 && b == 32 ) ||
	  ( a == 16 && b == 35 ) ||
	  ( a == 16 && b == 36 ) ||
	  ( a == 16 && b == 38 ) ||
	  ( a == 17 && b == 23 ) ||
	  ( a == 17 && b == 26 ) ||
	  ( a == 17 && b == 29 ) ||
	  ( a == 17 && b == 30 ) ||
	  ( a == 17 && b == 36 ) ||
	  ( a == 17 && b == 39 ) ||
	  ( a == 18 && b == 21 ) ||
	  ( a == 18 && b == 22 ) ||
	  ( a == 18 && b == 25 ) ||
	  ( a == 18 && b == 31 ) ||
	  ( a == 18 && b == 35 ) ||
	  ( a == 18 && b == 43 ) ||
	  ( a == 19 && b == 20 ) ||
	  ( a == 19 && b == 23 ) ||
	  ( a == 19 && b == 24 ) ||
	  ( a == 19 && b == 29 ) ||
	  ( a == 19 && b == 37 ) ||
	  ( a == 19 && b == 41 )  )
	return -sqr(-1 + Nc2)/(16.*Nc2);
      if( ( a == 0 && b == 28 ) ||
	  ( a == 0 && b == 30 ) ||
	  ( a == 0 && b == 34 ) ||
	  ( a == 0 && b == 35 ) ||
	  ( a == 0 && b == 40 ) ||
	  ( a == 0 && b == 41 ) ||
	  ( a == 1 && b == 28 ) ||
	  ( a == 1 && b == 29 ) ||
	  ( a == 1 && b == 34 ) ||
	  ( a == 1 && b == 36 ) ||
	  ( a == 1 && b == 42 ) ||
	  ( a == 1 && b == 43 ) ||
	  ( a == 2 && b == 30 ) ||
	  ( a == 2 && b == 31 ) ||
	  ( a == 2 && b == 36 ) ||
	  ( a == 2 && b == 37 ) ||
	  ( a == 2 && b == 40 ) ||
	  ( a == 2 && b == 42 ) ||
	  ( a == 3 && b == 22 ) ||
	  ( a == 3 && b == 24 ) ||
	  ( a == 3 && b == 32 ) ||
	  ( a == 3 && b == 33 ) ||
	  ( a == 3 && b == 38 ) ||
	  ( a == 3 && b == 39 ) ||
	  ( a == 4 && b == 22 ) ||
	  ( a == 4 && b == 23 ) ||
	  ( a == 4 && b == 32 ) ||
	  ( a == 4 && b == 37 ) ||
	  ( a == 4 && b == 42 ) ||
	  ( a == 4 && b == 43 ) ||
	  ( a == 5 && b == 24 ) ||
	  ( a == 5 && b == 25 ) ||
	  ( a == 5 && b == 36 ) ||
	  ( a == 5 && b == 37 ) ||
	  ( a == 5 && b == 38 ) ||
	  ( a == 5 && b == 43 ) ||
	  ( a == 6 && b == 20 ) ||
	  ( a == 6 && b == 25 ) ||
	  ( a == 6 && b == 26 ) ||
	  ( a == 6 && b == 27 ) ||
	  ( a == 6 && b == 38 ) ||
	  ( a == 6 && b == 39 ) ||
	  ( a == 7 && b == 20 ) ||
	  ( a == 7 && b == 21 ) ||
	  ( a == 7 && b == 26 ) ||
	  ( a == 7 && b == 31 ) ||
	  ( a == 7 && b == 40 ) ||
	  ( a == 7 && b == 41 ) ||
	  ( a == 8 && b == 24 ) ||
	  ( a == 8 && b == 25 ) ||
	  ( a == 8 && b == 30 ) ||
	  ( a == 8 && b == 31 ) ||
	  ( a == 8 && b == 39 ) ||
	  ( a == 8 && b == 41 ) ||
	  ( a == 9 && b == 21 ) ||
	  ( a == 9 && b == 23 ) ||
	  ( a == 9 && b == 26 ) ||
	  ( a == 9 && b == 27 ) ||
	  ( a == 9 && b == 32 ) ||
	  ( a == 9 && b == 33 ) ||
	  ( a == 10 && b == 20 ) ||
	  ( a == 10 && b == 21 ) ||
	  ( a == 10 && b == 27 ) ||
	  ( a == 10 && b == 29 ) ||
	  ( a == 10 && b == 34 ) ||
	  ( a == 10 && b == 35 ) ||
	  ( a == 11 && b == 22 ) ||
	  ( a == 11 && b == 23 ) ||
	  ( a == 11 && b == 28 ) ||
	  ( a == 11 && b == 29 ) ||
	  ( a == 11 && b == 33 ) ||
	  ( a == 11 && b == 35 ) ||
	  ( a == 12 && b == 23 ) ||
	  ( a == 12 && b == 25 ) ||
	  ( a == 12 && b == 27 ) ||
	  ( a == 12 && b == 30 ) ||
	  ( a == 12 && b == 35 ) ||
	  ( a == 12 && b == 37 ) ||
	  ( a == 13 && b == 23 ) ||
	  ( a == 13 && b == 25 ) ||
	  ( a == 13 && b == 26 ) ||
	  ( a == 13 && b == 28 ) ||
	  ( a == 13 && b == 41 ) ||
	  ( a == 13 && b == 43 ) ||
	  ( a == 14 && b == 21 ) ||
	  ( a == 14 && b == 24 ) ||
	  ( a == 14 && b == 29 ) ||
	  ( a == 14 && b == 31 ) ||
	  ( a == 14 && b == 33 ) ||
	  ( a == 14 && b == 36 ) ||
	  ( a == 15 && b == 21 ) ||
	  ( a == 15 && b == 24 ) ||
	  ( a == 15 && b == 32 ) ||
	  ( a == 15 && b == 34 ) ||
	  ( a == 15 && b == 41 ) ||
	  ( a == 15 && b == 43 ) ||
	  ( a == 16 && b == 20 ) ||
	  ( a == 16 && b == 22 ) ||
	  ( a == 16 && b == 29 ) ||
	  ( a == 16 && b == 31 ) ||
	  ( a == 16 && b == 39 ) ||
	  ( a == 16 && b == 42 ) ||
	  ( a == 17 && b == 20 ) ||
	  ( a == 17 && b == 22 ) ||
	  ( a == 17 && b == 35 ) ||
	  ( a == 17 && b == 37 ) ||
	  ( a == 17 && b == 38 ) ||
	  ( a == 17 && b == 40 ) ||
	  ( a == 18 && b == 27 ) ||
	  ( a == 18 && b == 30 ) ||
	  ( a == 18 && b == 32 ) ||
	  ( a == 18 && b == 34 ) ||
	  ( a == 18 && b == 39 ) ||
	  ( a == 18 && b == 42 ) ||
	  ( a == 19 && b == 26 ) ||
	  ( a == 19 && b == 28 ) ||
	  ( a == 19 && b == 33 ) ||
	  ( a == 19 && b == 36 ) ||
	  ( a == 19 && b == 38 ) ||
	  ( a == 19 && b == 40 )  )
	return (1 - 1/Nc2)/16.;
      if( ( a == 20 && b == 20 ) ||
	  ( a == 21 && b == 21 ) ||
	  ( a == 22 && b == 22 ) ||
	  ( a == 23 && b == 23 ) ||
	  ( a == 24 && b == 24 ) ||
	  ( a == 25 && b == 25 ) ||
	  ( a == 26 && b == 26 ) ||
	  ( a == 27 && b == 27 ) ||
	  ( a == 28 && b == 28 ) ||
	  ( a == 29 && b == 29 ) ||
	  ( a == 30 && b == 30 ) ||
	  ( a == 31 && b == 31 ) ||
	  ( a == 32 && b == 32 ) ||
	  ( a == 33 && b == 33 ) ||
	  ( a == 34 && b == 34 ) ||
	  ( a == 35 && b == 35 ) ||
	  ( a == 36 && b == 36 ) ||
	  ( a == 37 && b == 37 ) ||
	  ( a == 38 && b == 38 ) ||
	  ( a == 39 && b == 39 ) ||
	  ( a == 40 && b == 40 ) ||
	  ( a == 41 && b == 41 ) ||
	  ( a == 42 && b == 42 ) ||
	  ( a == 43 && b == 43 )  )
	return (4 - 10*Nc2 + 10*Nc4 - 5*Nc6 + Nc8)/(32.*Nc3);
      if( ( a == 20 && b == 21 ) ||
	  ( a == 20 && b == 22 ) ||
	  ( a == 20 && b == 26 ) ||
	  ( a == 20 && b == 29 ) ||
	  ( a == 20 && b == 38 ) ||
	  ( a == 21 && b == 24 ) ||
	  ( a == 21 && b == 27 ) ||
	  ( a == 21 && b == 31 ) ||
	  ( a == 21 && b == 32 ) ||
	  ( a == 22 && b == 23 ) ||
	  ( a == 22 && b == 32 ) ||
	  ( a == 22 && b == 35 ) ||
	  ( a == 22 && b == 39 ) ||
	  ( a == 23 && b == 25 ) ||
	  ( a == 23 && b == 26 ) ||
	  ( a == 23 && b == 33 ) ||
	  ( a == 23 && b == 37 ) ||
	  ( a == 24 && b == 25 ) ||
	  ( a == 24 && b == 33 ) ||
	  ( a == 24 && b == 38 ) ||
	  ( a == 24 && b == 41 ) ||
	  ( a == 25 && b == 27 ) ||
	  ( a == 25 && b == 39 ) ||
	  ( a == 25 && b == 43 ) ||
	  ( a == 26 && b == 27 ) ||
	  ( a == 26 && b == 28 ) ||
	  ( a == 26 && b == 40 ) ||
	  ( a == 27 && b == 30 ) ||
	  ( a == 27 && b == 34 ) ||
	  ( a == 28 && b == 29 ) ||
	  ( a == 28 && b == 33 ) ||
	  ( a == 28 && b == 34 ) ||
	  ( a == 28 && b == 41 ) ||
	  ( a == 29 && b == 31 ) ||
	  ( a == 29 && b == 35 ) ||
	  ( a == 29 && b == 36 ) ||
	  ( a == 30 && b == 31 ) ||
	  ( a == 30 && b == 35 ) ||
	  ( a == 30 && b == 39 ) ||
	  ( a == 30 && b == 40 ) ||
	  ( a == 31 && b == 41 ) ||
	  ( a == 31 && b == 42 ) ||
	  ( a == 32 && b == 33 ) ||
	  ( a == 32 && b == 34 ) ||
	  ( a == 32 && b == 42 ) ||
	  ( a == 33 && b == 36 ) ||
	  ( a == 34 && b == 35 ) ||
	  ( a == 34 && b == 43 ) ||
	  ( a == 35 && b == 37 ) ||
	  ( a == 36 && b == 37 ) ||
	  ( a == 36 && b == 38 ) ||
	  ( a == 36 && b == 42 ) ||
	  ( a == 37 && b == 40 ) ||
	  ( a == 37 && b == 43 ) ||
	  ( a == 38 && b == 39 ) ||
	  ( a == 38 && b == 40 ) ||
	  ( a == 39 && b == 42 ) ||
	  ( a == 40 && b == 41 ) ||
	  ( a == 41 && b == 43 ) ||
	  ( a == 42 && b == 43 )  )
	return -(-4 + 7*Nc2 - 4*Nc4 + Nc6)/(32.*Nc3);
      if( ( a == 20 && b == 23 ) ||
	  ( a == 20 && b == 24 ) ||
	  ( a == 20 && b == 28 ) ||
	  ( a == 20 && b == 32 ) ||
	  ( a == 20 && b == 36 ) ||
	  ( a == 21 && b == 22 ) ||
	  ( a == 21 && b == 25 ) ||
	  ( a == 21 && b == 30 ) ||
	  ( a == 21 && b == 38 ) ||
	  ( a == 21 && b == 42 ) ||
	  ( a == 22 && b == 25 ) ||
	  ( a == 22 && b == 26 ) ||
	  ( a == 22 && b == 30 ) ||
	  ( a == 22 && b == 34 ) ||
	  ( a == 23 && b == 24 ) ||
	  ( a == 23 && b == 36 ) ||
	  ( a == 23 && b == 39 ) ||
	  ( a == 23 && b == 40 ) ||
	  ( a == 24 && b == 27 ) ||
	  ( a == 24 && b == 28 ) ||
	  ( a == 24 && b == 40 ) ||
	  ( a == 25 && b == 33 ) ||
	  ( a == 25 && b == 34 ) ||
	  ( a == 25 && b == 42 ) ||
	  ( a == 26 && b == 29 ) ||
	  ( a == 26 && b == 30 ) ||
	  ( a == 26 && b == 34 ) ||
	  ( a == 26 && b == 37 ) ||
	  ( a == 27 && b == 28 ) ||
	  ( a == 27 && b == 31 ) ||
	  ( a == 27 && b == 40 ) ||
	  ( a == 27 && b == 43 ) ||
	  ( a == 28 && b == 31 ) ||
	  ( a == 28 && b == 32 ) ||
	  ( a == 29 && b == 30 ) ||
	  ( a == 29 && b == 37 ) ||
	  ( a == 29 && b == 38 ) ||
	  ( a == 29 && b == 41 ) ||
	  ( a == 30 && b == 38 ) ||
	  ( a == 31 && b == 32 ) ||
	  ( a == 31 && b == 35 ) ||
	  ( a == 31 && b == 43 ) ||
	  ( a == 32 && b == 35 ) ||
	  ( a == 32 && b == 36 ) ||
	  ( a == 33 && b == 34 ) ||
	  ( a == 33 && b == 37 ) ||
	  ( a == 33 && b == 41 ) ||
	  ( a == 33 && b == 42 ) ||
	  ( a == 34 && b == 37 ) ||
	  ( a == 35 && b == 36 ) ||
	  ( a == 35 && b == 39 ) ||
	  ( a == 35 && b == 43 ) ||
	  ( a == 36 && b == 39 ) ||
	  ( a == 37 && b == 41 ) ||
	  ( a == 38 && b == 41 ) ||
	  ( a == 38 && b == 42 ) ||
	  ( a == 39 && b == 40 ) ||
	  ( a == 39 && b == 43 ) ||
	  ( a == 40 && b == 43 ) ||
	  ( a == 41 && b == 42 )  )
	return (2 - 3*Nc2 + Nc4)/(16.*Nc3);
      if( ( a == 20 && b == 25 ) ||
	  ( a == 20 && b == 34 ) ||
	  ( a == 20 && b == 37 ) ||
	  ( a == 20 && b == 41 ) ||
	  ( a == 20 && b == 42 ) ||
	  ( a == 21 && b == 23 ) ||
	  ( a == 21 && b == 35 ) ||
	  ( a == 21 && b == 36 ) ||
	  ( a == 21 && b == 40 ) ||
	  ( a == 21 && b == 43 ) ||
	  ( a == 22 && b == 24 ) ||
	  ( a == 22 && b == 28 ) ||
	  ( a == 22 && b == 31 ) ||
	  ( a == 22 && b == 40 ) ||
	  ( a == 22 && b == 43 ) ||
	  ( a == 23 && b == 29 ) ||
	  ( a == 23 && b == 30 ) ||
	  ( a == 23 && b == 41 ) ||
	  ( a == 23 && b == 42 ) ||
	  ( a == 24 && b == 29 ) ||
	  ( a == 24 && b == 30 ) ||
	  ( a == 24 && b == 34 ) ||
	  ( a == 24 && b == 37 ) ||
	  ( a == 25 && b == 28 ) ||
	  ( a == 25 && b == 31 ) ||
	  ( a == 25 && b == 35 ) ||
	  ( a == 25 && b == 36 ) ||
	  ( a == 26 && b == 31 ) ||
	  ( a == 26 && b == 32 ) ||
	  ( a == 26 && b == 36 ) ||
	  ( a == 26 && b == 39 ) ||
	  ( a == 26 && b == 43 ) ||
	  ( a == 27 && b == 29 ) ||
	  ( a == 27 && b == 33 ) ||
	  ( a == 27 && b == 37 ) ||
	  ( a == 27 && b == 38 ) ||
	  ( a == 27 && b == 42 ) ||
	  ( a == 28 && b == 30 ) ||
	  ( a == 28 && b == 38 ) ||
	  ( a == 28 && b == 42 ) ||
	  ( a == 29 && b == 39 ) ||
	  ( a == 29 && b == 43 ) ||
	  ( a == 30 && b == 32 ) ||
	  ( a == 30 && b == 36 ) ||
	  ( a == 31 && b == 33 ) ||
	  ( a == 31 && b == 37 ) ||
	  ( a == 32 && b == 37 ) ||
	  ( a == 32 && b == 38 ) ||
	  ( a == 32 && b == 41 ) ||
	  ( a == 33 && b == 35 ) ||
	  ( a == 33 && b == 39 ) ||
	  ( a == 33 && b == 40 ) ||
	  ( a == 34 && b == 36 ) ||
	  ( a == 34 && b == 39 ) ||
	  ( a == 34 && b == 40 ) ||
	  ( a == 35 && b == 38 ) ||
	  ( a == 35 && b == 41 ) ||
	  ( a == 38 && b == 43 ) ||
	  ( a == 39 && b == 41 ) ||
	  ( a == 40 && b == 42 )  )
	return (4 - 3*Nc2 - 2*Nc4 + Nc6)/(32.*Nc3);
      if( ( a == 20 && b == 27 ) ||
	  ( a == 20 && b == 31 ) ||
	  ( a == 20 && b == 35 ) ||
	  ( a == 20 && b == 39 ) ||
	  ( a == 20 && b == 40 ) ||
	  ( a == 21 && b == 26 ) ||
	  ( a == 21 && b == 29 ) ||
	  ( a == 21 && b == 33 ) ||
	  ( a == 21 && b == 34 ) ||
	  ( a == 21 && b == 41 ) ||
	  ( a == 22 && b == 29 ) ||
	  ( a == 22 && b == 33 ) ||
	  ( a == 22 && b == 37 ) ||
	  ( a == 22 && b == 38 ) ||
	  ( a == 22 && b == 42 ) ||
	  ( a == 23 && b == 27 ) ||
	  ( a == 23 && b == 28 ) ||
	  ( a == 23 && b == 32 ) ||
	  ( a == 23 && b == 35 ) ||
	  ( a == 23 && b == 43 ) ||
	  ( a == 24 && b == 31 ) ||
	  ( a == 24 && b == 32 ) ||
	  ( a == 24 && b == 36 ) ||
	  ( a == 24 && b == 39 ) ||
	  ( a == 24 && b == 43 ) ||
	  ( a == 25 && b == 26 ) ||
	  ( a == 25 && b == 30 ) ||
	  ( a == 25 && b == 37 ) ||
	  ( a == 25 && b == 38 ) ||
	  ( a == 25 && b == 41 ) ||
	  ( a == 26 && b == 33 ) ||
	  ( a == 26 && b == 38 ) ||
	  ( a == 26 && b == 41 ) ||
	  ( a == 27 && b == 32 ) ||
	  ( a == 27 && b == 35 ) ||
	  ( a == 27 && b == 39 ) ||
	  ( a == 28 && b == 35 ) ||
	  ( a == 28 && b == 36 ) ||
	  ( a == 28 && b == 40 ) ||
	  ( a == 28 && b == 43 ) ||
	  ( a == 29 && b == 33 ) ||
	  ( a == 29 && b == 34 ) ||
	  ( a == 29 && b == 42 ) ||
	  ( a == 30 && b == 34 ) ||
	  ( a == 30 && b == 37 ) ||
	  ( a == 30 && b == 41 ) ||
	  ( a == 30 && b == 42 ) ||
	  ( a == 31 && b == 36 ) ||
	  ( a == 31 && b == 39 ) ||
	  ( a == 31 && b == 40 ) ||
	  ( a == 32 && b == 39 ) ||
	  ( a == 32 && b == 43 ) ||
	  ( a == 33 && b == 38 ) ||
	  ( a == 34 && b == 41 ) ||
	  ( a == 34 && b == 42 ) ||
	  ( a == 35 && b == 40 ) ||
	  ( a == 36 && b == 40 ) ||
	  ( a == 36 && b == 43 ) ||
	  ( a == 37 && b == 38 ) ||
	  ( a == 37 && b == 42 )  )
	return -(-1 + Nc2)/(8.*Nc3);
      if( ( a == 20 && b == 30 ) ||
	  ( a == 20 && b == 33 ) ||
	  ( a == 21 && b == 28 ) ||
	  ( a == 21 && b == 39 ) ||
	  ( a == 22 && b == 27 ) ||
	  ( a == 22 && b == 36 ) ||
	  ( a == 23 && b == 34 ) ||
	  ( a == 23 && b == 38 ) ||
	  ( a == 24 && b == 26 ) ||
	  ( a == 24 && b == 42 ) ||
	  ( a == 25 && b == 32 ) ||
	  ( a == 25 && b == 40 ) ||
	  ( a == 26 && b == 35 ) ||
	  ( a == 27 && b == 41 ) ||
	  ( a == 28 && b == 37 ) ||
	  ( a == 29 && b == 32 ) ||
	  ( a == 29 && b == 40 ) ||
	  ( a == 30 && b == 43 ) ||
	  ( a == 31 && b == 34 ) ||
	  ( a == 31 && b == 38 ) ||
	  ( a == 33 && b == 43 ) ||
	  ( a == 35 && b == 42 ) ||
	  ( a == 36 && b == 41 ) ||
	  ( a == 37 && b == 39 )  )
	return (4 - 5*Nc2 + Nc4)/(32.*Nc3);
      if( ( a == 20 && b == 43 ) ||
	  ( a == 21 && b == 37 ) ||
	  ( a == 22 && b == 41 ) ||
	  ( a == 23 && b == 31 ) ||
	  ( a == 24 && b == 35 ) ||
	  ( a == 25 && b == 29 ) ||
	  ( a == 26 && b == 42 ) ||
	  ( a == 27 && b == 36 ) ||
	  ( a == 28 && b == 39 ) ||
	  ( a == 30 && b == 33 ) ||
	  ( a == 32 && b == 40 ) ||
	  ( a == 34 && b == 38 )  )
	return -(-1 + Nc4)/(8.*Nc3);
    }

    if ( basis == id33bar888 ) {
      if( ( a == 0 && b == 0 ) ||
	  ( a == 1 && b == 1 )  )
	return (2 - 3*Nc2 + Nc4)/8.;
      if( a == 0 && b == 1  )
	return (1 - Nc2)/4.;
      if( ( a == 0 && b == 2 ) ||
	  ( a == 0 && b == 3 ) ||
	  ( a == 0 && b == 4 ) ||
	  ( a == 1 && b == 2 ) ||
	  ( a == 1 && b == 3 ) ||
	  ( a == 1 && b == 4 )  )
	return 0;
      if( ( a == 0 && b == 5 ) ||
	  ( a == 0 && b == 8 ) ||
	  ( a == 0 && b == 9 ) ||
	  ( a == 1 && b == 6 ) ||
	  ( a == 1 && b == 7 ) ||
	  ( a == 1 && b == 10 )  )
	return (2 - 3*Nc2 + Nc4)/(8.*Nc);
      if( ( a == 0 && b == 6 ) ||
	  ( a == 0 && b == 7 ) ||
	  ( a == 0 && b == 10 ) ||
	  ( a == 1 && b == 5 ) ||
	  ( a == 1 && b == 8 ) ||
	  ( a == 1 && b == 9 )  )
	return -(-1 + Nc2)/(4.*Nc);
      if( ( a == 2 && b == 2 ) ||
	  ( a == 3 && b == 3 ) ||
	  ( a == 4 && b == 4 )  )
	return sqr(-1 + Nc2)/8.;
      if( ( a == 2 && b == 3 ) ||
	  ( a == 2 && b == 4 ) ||
	  ( a == 3 && b == 4 )  )
	return (-1 + Nc2)/8.;
      if( ( a == 2 && b == 5 ) ||
	  ( a == 2 && b == 6 ) ||
	  ( a == 2 && b == 8 ) ||
	  ( a == 2 && b == 10 ) ||
	  ( a == 3 && b == 6 ) ||
	  ( a == 3 && b == 7 ) ||
	  ( a == 3 && b == 8 ) ||
	  ( a == 3 && b == 9 ) ||
	  ( a == 4 && b == 5 ) ||
	  ( a == 4 && b == 7 ) ||
	  ( a == 4 && b == 9 ) ||
	  ( a == 4 && b == 10 )  )
	return sqr(-1 + Nc2)/(8.*Nc);
      if( ( a == 2 && b == 7 ) ||
	  ( a == 2 && b == 9 ) ||
	  ( a == 3 && b == 5 ) ||
	  ( a == 3 && b == 10 ) ||
	  ( a == 4 && b == 6 ) ||
	  ( a == 4 && b == 8 )  )
	return -(-1 + Nc2)/(8.*Nc);
      if( ( a == 5 && b == 5 ) ||
	  ( a == 6 && b == 6 ) ||
	  ( a == 7 && b == 7 ) ||
	  ( a == 8 && b == 8 ) ||
	  ( a == 9 && b == 9 ) ||
	  ( a == 10 && b == 10 )  )
	return pow(-1 + Nc2,3.)/(8.*Nc2);
      if( ( a == 5 && b == 6 ) ||
	  ( a == 5 && b == 7 ) ||
	  ( a == 6 && b == 9 ) ||
	  ( a == 7 && b == 8 ) ||
	  ( a == 8 && b == 10 ) ||
	  ( a == 9 && b == 10 )  )
	return -sqr(-1 + Nc2)/(8.*Nc2);
      if( ( a == 5 && b == 8 ) ||
	  ( a == 5 && b == 9 ) ||
	  ( a == 6 && b == 7 ) ||
	  ( a == 6 && b == 10 ) ||
	  ( a == 7 && b == 10 ) ||
	  ( a == 8 && b == 9 )  )
	return 0.125 - 1/(8.*Nc2);
      if( ( a == 5 && b == 10 ) ||
	  ( a == 6 && b == 8 ) ||
	  ( a == 7 && b == 9 )  )
	return (-1 + Nc4)/(8.*Nc2);
    }

    if ( basis == id33bar33bar8 ) {
      if( ( a == 0 && b == 0 ) ||
	  ( a == 1 && b == 1 ) ||
	  ( a == 2 && b == 2 ) ||
	  ( a == 3 && b == 3 )  )
	return (Nc*(-1 + Nc2))/2.;
      if( ( a == 0 && b == 1 ) ||
	  ( a == 0 && b == 3 ) ||
	  ( a == 1 && b == 2 ) ||
	  ( a == 2 && b == 3 )  )
	return (-1 + Nc2)/2.;
      if( ( a == 0 && b == 2 ) ||
	  ( a == 1 && b == 3 )  )
	return 0;
    }

  } else {

    if ( a != b )
      return 0.;

    if ( basis == id88 ) {
      return Nc2/4.;
    }

    if ( basis == id33bar ) {
      return Nc;
    }

    if ( basis == id888 ) {
      return Nc3/8.;
    }

    if ( basis == id33bar8 ) {
      return Nc2/2.;
    }

    if ( basis == id8888 ) {
      return Nc4/16.;
    }

    if ( basis == id33bar88 ) {
      return Nc3/4.;
    }

    if ( basis == id33bar33bar ) {
      return Nc2;
    }

    if ( basis == id88888 ) {
      return Nc5/32.;
    }

    if ( basis == id33bar888 ) {
      return Nc4/8.;
    }

    if ( basis == id33bar33bar8 ) {
      return Nc3/2.;
    }

  }

  throw Exception() << "SimpleColourBasis2::scalarProduct(): Cannot handle colour configuration" << Exception::runerror;

}

double SimpleColourBasis2::tMatrixElement(size_t i, size_t a, 
					  size_t b,
					  const vector<PDT::Colour>&,
					  const vector<PDT::Colour>& basis,
#ifndef NDEBUG
					  size_t k, size_t l,
#else
					  size_t , size_t ,
#endif
					  const map<size_t,size_t>& dict) const {
  // Check indices k and l
  assert( k == i );
  assert( l == basis.size() );
  // Check that dict is the standardMap
  assert( dict.size()+1 == basis.size() );
  map<size_t,size_t>::const_iterator tmp;
  for ( size_t ii = 0; ii < basis.size(); ii++ )
    if ( ii != i ) {
      tmp = dict.find(ii);
      assert( tmp != dict.end() ); 
      assert( tmp->second == ii );
    }

  if ( id33bar.empty() )
    makeIds();

  if ( basis == id88 ) {
    if(i == 0 && a == 0 && b == 0) return -1;
    if(i == 0 && a == 1 && b == 0) return 1;
    if(i == 1 && a == 0 && b == 0) return 1;
    if(i == 1 && a == 1 && b == 0) return -1;
    return 0.;
  }

  if ( basis == id33bar ) {
    if(i == 0 && a == 0 && b == 0) return 1;
    if(i == 1 && a == 0 && b == 0) return -1;
    return 0.;
  }

  if ( basis == id888 ) {
    if(i == 0 && a == 3 && b == 0) return -1;
    if(i == 0 && a == 5 && b == 1) return -1;
    if(i == 0 && a == 7 && b == 0) return 1;
    if(i == 0 && a == 8 && b == 1) return 1;
    if(i == 1 && a == 4 && b == 0) return 1;
    if(i == 1 && a == 5 && b == 1) return 1;
    if(i == 1 && a == 6 && b == 1) return -1;
    if(i == 1 && a == 7 && b == 0) return -1;
    if(i == 2 && a == 3 && b == 0) return 1;
    if(i == 2 && a == 4 && b == 0) return -1;
    if(i == 2 && a == 6 && b == 1) return 1;
    if(i == 2 && a == 8 && b == 1) return -1;
    return 0.;
  }

  if ( basis == id33bar8 ) {
    if(i == 0 && a == 2 && b == 0) return 1;
    if(i == 1 && a == 1 && b == 0) return -1;
    if(i == 2 && a == 1 && b == 0) return 1;
    if(i == 2 && a == 2 && b == 0) return -1;
    return 0.;
  }

  if ( basis == id8888 ) {
    if(i == 0 && a == 2 && b == 2) return -1;
    if(i == 0 && a == 5 && b == 1) return -1;
    if(i == 0 && a == 8 && b == 0) return -1;
    if(i == 0 && a == 9 && b == 2) return 1;
    if(i == 0 && a == 10 && b == 1) return 1;
    if(i == 0 && a == 11 && b == 0) return 1;
    if(i == 0 && a == 20 && b == 3) return -1;
    if(i == 0 && a == 22 && b == 4) return -1;
    if(i == 0 && a == 26 && b == 5) return -1;
    if(i == 0 && a == 28 && b == 6) return -1;
    if(i == 0 && a == 32 && b == 7) return -1;
    if(i == 0 && a == 34 && b == 8) return -1;
    if(i == 0 && a == 38 && b == 3) return 1;
    if(i == 0 && a == 39 && b == 4) return 1;
    if(i == 0 && a == 40 && b == 5) return 1;
    if(i == 0 && a == 41 && b == 6) return 1;
    if(i == 0 && a == 42 && b == 7) return 1;
    if(i == 0 && a == 43 && b == 8) return 1;
    if(i == 1 && a == 2 && b == 2) return 1;
    if(i == 1 && a == 9 && b == 2) return -1;
    if(i == 1 && a == 13 && b == 0) return -1;
    if(i == 1 && a == 15 && b == 1) return -1;
    if(i == 1 && a == 16 && b == 0) return 1;
    if(i == 1 && a == 17 && b == 1) return 1;
    if(i == 1 && a == 24 && b == 3) return 1;
    if(i == 1 && a == 25 && b == 4) return 1;
    if(i == 1 && a == 27 && b == 5) return 1;
    if(i == 1 && a == 28 && b == 6) return 1;
    if(i == 1 && a == 29 && b == 6) return -1;
    if(i == 1 && a == 30 && b == 5) return -1;
    if(i == 1 && a == 33 && b == 7) return 1;
    if(i == 1 && a == 34 && b == 8) return 1;
    if(i == 1 && a == 35 && b == 8) return -1;
    if(i == 1 && a == 36 && b == 7) return -1;
    if(i == 1 && a == 38 && b == 3) return -1;
    if(i == 1 && a == 39 && b == 4) return -1;
    if(i == 2 && a == 5 && b == 1) return 1;
    if(i == 2 && a == 10 && b == 1) return -1;
    if(i == 2 && a == 13 && b == 0) return 1;
    if(i == 2 && a == 16 && b == 0) return -1;
    if(i == 2 && a == 18 && b == 2) return -1;
    if(i == 2 && a == 19 && b == 2) return 1;
    if(i == 2 && a == 21 && b == 3) return 1;
    if(i == 2 && a == 22 && b == 4) return 1;
    if(i == 2 && a == 23 && b == 4) return -1;
    if(i == 2 && a == 24 && b == 3) return -1;
    if(i == 2 && a == 30 && b == 5) return 1;
    if(i == 2 && a == 31 && b == 6) return 1;
    if(i == 2 && a == 32 && b == 7) return 1;
    if(i == 2 && a == 33 && b == 7) return -1;
    if(i == 2 && a == 35 && b == 8) return 1;
    if(i == 2 && a == 37 && b == 8) return -1;
    if(i == 2 && a == 40 && b == 5) return -1;
    if(i == 2 && a == 41 && b == 6) return -1;
    if(i == 3 && a == 8 && b == 0) return 1;
    if(i == 3 && a == 11 && b == 0) return -1;
    if(i == 3 && a == 15 && b == 1) return 1;
    if(i == 3 && a == 17 && b == 1) return -1;
    if(i == 3 && a == 18 && b == 2) return 1;
    if(i == 3 && a == 19 && b == 2) return -1;
    if(i == 3 && a == 20 && b == 3) return 1;
    if(i == 3 && a == 21 && b == 3) return -1;
    if(i == 3 && a == 23 && b == 4) return 1;
    if(i == 3 && a == 25 && b == 4) return -1;
    if(i == 3 && a == 26 && b == 5) return 1;
    if(i == 3 && a == 27 && b == 5) return -1;
    if(i == 3 && a == 29 && b == 6) return 1;
    if(i == 3 && a == 31 && b == 6) return -1;
    if(i == 3 && a == 36 && b == 7) return 1;
    if(i == 3 && a == 37 && b == 8) return 1;
    if(i == 3 && a == 42 && b == 7) return -1;
    if(i == 3 && a == 43 && b == 8) return -1;
    return 0.;
  }

  if ( basis == id33bar88 ) {
    if(i == 0 && a == 4 && b == 0) return 1;
    if(i == 0 && a == 9 && b == 1) return 1;
    if(i == 0 && a == 10 && b == 2) return 1;
    if(i == 1 && a == 4 && b == 0) return -1;
    if(i == 1 && a == 5 && b == 1) return -1;
    if(i == 1 && a == 7 && b == 2) return -1;
    if(i == 2 && a == 0 && b == 0) return -1;
    if(i == 2 && a == 1 && b == 0) return 1;
    if(i == 2 && a == 6 && b == 1) return 1;
    if(i == 2 && a == 7 && b == 2) return 1;
    if(i == 2 && a == 8 && b == 2) return -1;
    if(i == 2 && a == 9 && b == 1) return -1;
    if(i == 3 && a == 0 && b == 0) return 1;
    if(i == 3 && a == 1 && b == 0) return -1;
    if(i == 3 && a == 5 && b == 1) return 1;
    if(i == 3 && a == 6 && b == 1) return -1;
    if(i == 3 && a == 8 && b == 2) return 1;
    if(i == 3 && a == 10 && b == 2) return -1;
    return 0.;
  }

  if ( basis == id33bar33bar ) {
    if(i == 0 && a == 0 && b == 0) return 1;
    if(i == 0 && a == 1 && b == 1) return 1;
    if(i == 1 && a == 1 && b == 1) return -1;
    if(i == 1 && a == 2 && b == 0) return -1;
    if(i == 2 && a == 2 && b == 0) return 1;
    if(i == 2 && a == 3 && b == 1) return 1;
    if(i == 3 && a == 0 && b == 0) return -1;
    if(i == 3 && a == 3 && b == 1) return -1;
    return 0.;
  }

  throw Exception() << "SimpleColourBasis2::tMatrixElement(): Cannot handle colour configuration" << Exception::runerror;

  return 0.;

}

bool SimpleColourBasis2::colourConnected(const cPDVector& sub,
					const vector<PDT::Colour>& basis,
					const pair<int,bool>& i, 
					const pair<int,bool>& j, 
					size_t a) const {

  if ( id33bar.empty() )
    makeIds();

  // translate process to basis ids
  map<cPDVector,map<size_t,size_t> >::const_iterator trans
    = indexMap().find(sub);
  assert(trans != indexMap().end());

  int idColoured = i.second ? j.first : i.first;
  idColoured = trans->second.find(idColoured)->second;
  int idAntiColoured = i.second ? i.first : j.first;
  idAntiColoured = trans->second.find(idAntiColoured)->second;

  if ( basis == id88 ) {
    return
      a == 0 &&
      ((idColoured == 0 && idAntiColoured == 1) ||
       (idColoured == 1 && idAntiColoured == 0));
  }

  if ( basis == id33bar ) {
    return
      a == 0 &&
      (idColoured == 0 && idAntiColoured == 1);
  }

  if ( basis == id888 ) {
    return
      (a == 0 &&
       ((idColoured == 0 && idAntiColoured == 1) ||
	(idColoured == 1 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 0))) ||
      (a == 1 &&
       ((idColoured == 0 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 1) ||
	(idColoured == 1 && idAntiColoured == 0)));
  }

  if ( basis == id33bar8 ) {
    return
      a == 0 &&
      ((idColoured == 0 && idAntiColoured == 2) ||
       (idColoured == 2 && idAntiColoured == 1));
  }

  if ( basis == id8888 ) {
    return
      (a == 0 &&
       ((idColoured == 0 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 0) ||
	(idColoured == 1 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 1))) ||
      (a == 1 &&
       ((idColoured == 0 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 0) ||
	(idColoured == 1 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 1))) ||
      (a == 2 &&
       ((idColoured == 0 && idAntiColoured == 1) ||
	(idColoured == 1 && idAntiColoured == 0) ||
	(idColoured == 2 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 2))) ||
      (a == 3 &&
       ((idColoured == 0 && idAntiColoured == 1) ||
	(idColoured == 1 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 0))) ||
      (a == 4 &&
       ((idColoured == 0 && idAntiColoured == 1) ||
	(idColoured == 1 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 0))) ||
      (a == 5 &&
       ((idColoured == 0 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 1) ||
	(idColoured == 1 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 0))) ||
      (a == 6 &&
       ((idColoured == 0 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 1) ||
	(idColoured == 1 && idAntiColoured == 0))) ||
      (a == 7 &&
       ((idColoured == 0 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 1) ||
	(idColoured == 1 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 0))) ||
      (a == 8 &&
       ((idColoured == 0 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 1) ||
	(idColoured == 1 && idAntiColoured == 0)));
  }

  if ( basis == id33bar88 ) {
    return
      (a == 0 &&
       ((idColoured == 2 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 2) ||
	(idColoured == 0 && idAntiColoured == 1))) ||
      (a == 1 &&
       ((idColoured == 0 && idAntiColoured == 2) ||
	(idColoured == 3 && idAntiColoured == 1) ||
	(idColoured == 2 && idAntiColoured == 3))) ||
      (a == 2 &&
       ((idColoured == 0 && idAntiColoured == 3) ||
	(idColoured == 2 && idAntiColoured == 1) ||
	(idColoured == 3 && idAntiColoured == 2)));
  }

  if ( basis == id33bar33bar ) {
    return
      (a == 0 &&
       ((idColoured == 0 && idAntiColoured == 3) ||
	(idColoured == 2 && idAntiColoured == 1))) ||
      (a == 1 &&
       ((idColoured == 0 && idAntiColoured == 1) ||
	(idColoured == 2 && idAntiColoured == 3)));
  }

  if ( basis == id88888 ) {
    return
      (a == 0 &&
       ((idColoured == 3 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 3) ||
	(idColoured == 0 && idAntiColoured == 1) ||
	(idColoured == 1 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 0))) ||
      (a == 1 &&
       ((idColoured == 2 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 2) ||
	(idColoured == 0 && idAntiColoured == 1) ||
	(idColoured == 1 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 0))) ||
      (a == 2 &&
       ((idColoured == 2 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 2) ||
	(idColoured == 0 && idAntiColoured == 1) ||
	(idColoured == 1 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 0))) ||
      (a == 3 &&
       ((idColoured == 3 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 3) ||
	(idColoured == 0 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 1) ||
	(idColoured == 1 && idAntiColoured == 0))) ||
      (a == 4 &&
       ((idColoured == 1 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 1) ||
	(idColoured == 0 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 0))) ||
      (a == 5 &&
       ((idColoured == 1 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 1) ||
	(idColoured == 0 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 0))) ||
      (a == 6 &&
       ((idColoured == 2 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 2) ||
	(idColoured == 0 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 1) ||
	(idColoured == 1 && idAntiColoured == 0))) ||
      (a == 7 &&
       ((idColoured == 1 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 1) ||
	(idColoured == 0 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 0))) ||
      (a == 8 &&
       ((idColoured == 1 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 1) ||
	(idColoured == 0 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 0))) ||
      (a == 9 &&
       ((idColoured == 2 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 2) ||
	(idColoured == 0 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 1) ||
	(idColoured == 1 && idAntiColoured == 0))) ||
      (a == 10 &&
       ((idColoured == 1 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 1) ||
	(idColoured == 0 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 0))) ||
      (a == 11 &&
       ((idColoured == 1 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 1) ||
	(idColoured == 0 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 0))) ||
      (a == 12 &&
       ((idColoured == 0 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 0) ||
	(idColoured == 1 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 1))) ||
      (a == 13 &&
       ((idColoured == 0 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 0) ||
	(idColoured == 1 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 1))) ||
      (a == 14 &&
       ((idColoured == 0 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 0) ||
	(idColoured == 1 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 1))) ||
      (a == 15 &&
       ((idColoured == 0 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 0) ||
	(idColoured == 1 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 1))) ||
      (a == 16 &&
       ((idColoured == 0 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 0) ||
	(idColoured == 1 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 1))) ||
      (a == 17 &&
       ((idColoured == 0 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 0) ||
	(idColoured == 1 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 1))) ||
      (a == 18 &&
       ((idColoured == 0 && idAntiColoured == 1) ||
	(idColoured == 1 && idAntiColoured == 0) ||
	(idColoured == 2 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 2))) ||
      (a == 19 &&
       ((idColoured == 0 && idAntiColoured == 1) ||
	(idColoured == 1 && idAntiColoured == 0) ||
	(idColoured == 2 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 2))) ||
      (a == 20 &&
       ((idColoured == 0 && idAntiColoured == 1) ||
	(idColoured == 1 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 0))) ||
      (a == 21 &&
       ((idColoured == 0 && idAntiColoured == 1) ||
	(idColoured == 1 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 0))) ||
      (a == 22 &&
       ((idColoured == 0 && idAntiColoured == 1) ||
	(idColoured == 1 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 0))) ||
      (a == 23 &&
       ((idColoured == 0 && idAntiColoured == 1) ||
	(idColoured == 1 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 0))) ||
      (a == 24 &&
       ((idColoured == 0 && idAntiColoured == 1) ||
	(idColoured == 1 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 0))) ||
      (a == 25 &&
       ((idColoured == 0 && idAntiColoured == 1) ||
	(idColoured == 1 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 0))) ||
      (a == 26 &&
       ((idColoured == 0 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 1) ||
	(idColoured == 1 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 0))) ||
      (a == 27 &&
       ((idColoured == 0 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 1) ||
	(idColoured == 1 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 0))) ||
      (a == 28 &&
       ((idColoured == 0 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 1) ||
	(idColoured == 1 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 0))) ||
      (a == 29 &&
       ((idColoured == 0 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 1) ||
	(idColoured == 1 && idAntiColoured == 0))) ||
      (a == 30 &&
       ((idColoured == 0 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 1) ||
	(idColoured == 1 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 0))) ||
      (a == 31 &&
       ((idColoured == 0 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 1) ||
	(idColoured == 1 && idAntiColoured == 0))) ||
      (a == 32 &&
       ((idColoured == 0 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 1) ||
	(idColoured == 1 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 0))) ||
      (a == 33 &&
       ((idColoured == 0 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 1) ||
	(idColoured == 1 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 0))) ||
      (a == 34 &&
       ((idColoured == 0 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 1) ||
	(idColoured == 1 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 0))) ||
      (a == 35 &&
       ((idColoured == 0 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 1) ||
	(idColoured == 1 && idAntiColoured == 0))) ||
      (a == 36 &&
       ((idColoured == 0 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 1) ||
	(idColoured == 1 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 0))) ||
      (a == 37 &&
       ((idColoured == 0 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 1) ||
	(idColoured == 1 && idAntiColoured == 0))) ||
      (a == 38 &&
       ((idColoured == 0 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 1) ||
	(idColoured == 1 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 0))) ||
      (a == 39 &&
       ((idColoured == 0 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 1) ||
	(idColoured == 1 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 0))) ||
      (a == 40 &&
       ((idColoured == 0 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 1) ||
	(idColoured == 1 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 0))) ||
      (a == 41 &&
       ((idColoured == 0 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 1) ||
	(idColoured == 1 && idAntiColoured == 0))) ||
      (a == 42 &&
       ((idColoured == 0 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 1) ||
	(idColoured == 1 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 0))) ||
      (a == 43 &&
       ((idColoured == 0 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 1) ||
	(idColoured == 1 && idAntiColoured == 0)));
  }

  if ( basis == id33bar888 ) {
    return
      (a == 0 &&
       ((idColoured == 2 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 2) ||
	(idColoured == 0 && idAntiColoured == 1))) ||
      (a == 1 &&
       ((idColoured == 2 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 2) ||
	(idColoured == 0 && idAntiColoured == 1))) ||
      (a == 2 &&
       ((idColoured == 3 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 3) ||
	(idColoured == 0 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 1))) ||
      (a == 3 &&
       ((idColoured == 2 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 2) ||
	(idColoured == 0 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 1))) ||
      (a == 4 &&
       ((idColoured == 2 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 2) ||
	(idColoured == 0 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 1))) ||
      (a == 5 &&
       ((idColoured == 0 && idAntiColoured == 2) ||
	(idColoured == 4 && idAntiColoured == 1) ||
	(idColoured == 2 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 4))) ||
      (a == 6 &&
       ((idColoured == 0 && idAntiColoured == 2) ||
	(idColoured == 3 && idAntiColoured == 1) ||
	(idColoured == 2 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 3))) ||
      (a == 7 &&
       ((idColoured == 0 && idAntiColoured == 3) ||
	(idColoured == 4 && idAntiColoured == 1) ||
	(idColoured == 3 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 4))) ||
      (a == 8 &&
       ((idColoured == 0 && idAntiColoured == 3) ||
	(idColoured == 2 && idAntiColoured == 1) ||
	(idColoured == 3 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 2))) ||
      (a == 9 &&
       ((idColoured == 0 && idAntiColoured == 4) ||
	(idColoured == 3 && idAntiColoured == 1) ||
	(idColoured == 4 && idAntiColoured == 2) ||
	(idColoured == 2 && idAntiColoured == 3))) ||
      (a == 10 &&
       ((idColoured == 0 && idAntiColoured == 4) ||
	(idColoured == 2 && idAntiColoured == 1) ||
	(idColoured == 4 && idAntiColoured == 3) ||
	(idColoured == 3 && idAntiColoured == 2)));
  }

  if ( basis == id33bar33bar8 ) {
    return
      (a == 0 &&
       ((idColoured == 0 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 3) ||
	(idColoured == 2 && idAntiColoured == 1))) ||
      (a == 1 &&
       ((idColoured == 0 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 1) ||
	(idColoured == 2 && idAntiColoured == 3))) ||
      (a == 2 &&
       ((idColoured == 0 && idAntiColoured == 3) ||
	(idColoured == 2 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 1))) ||
      (a == 3 &&
       ((idColoured == 0 && idAntiColoured == 1) ||
	(idColoured == 2 && idAntiColoured == 4) ||
	(idColoured == 4 && idAntiColoured == 3)));
  }

  throw Exception() << "SimpleColourBasis2::colourConnected(): Cannot handle colour configuration" << Exception::runerror;

  return false;

}

map<size_t,vector<vector<size_t> > > SimpleColourBasis2::basisList(const vector<PDT::Colour>& basis) const {

  if ( id33bar.empty() )
    makeIds();

  map<size_t,vector<vector<size_t> > > blist;
  vector<vector<size_t> > structures;
  vector<size_t> structure;

  if ( basis == id88 ) {
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(1);
    structures.push_back(structure);
    blist[0] = structures;
    return blist;
  }

  if ( basis == id33bar ) {
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(1);
    structures.push_back(structure);
    blist[0] = structures;
    return blist;
  }

  if ( basis == id888 ) {
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(1);
    structure.push_back(2);
    structures.push_back(structure);
    blist[0] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(2);
    structure.push_back(1);
    structures.push_back(structure);
    blist[1] = structures;
    return blist;
  }

  if ( basis == id33bar8 ) {
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(2);
    structure.push_back(1);
    structures.push_back(structure);
    blist[0] = structures;
    return blist;
  }

  if ( basis == id8888 ) {
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(3);
    structures.push_back(structure);
    structure.clear();
    structure.push_back(1);
    structure.push_back(2);
    structures.push_back(structure);
    blist[0] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(2);
    structures.push_back(structure);
    structure.clear();
    structure.push_back(1);
    structure.push_back(3);
    structures.push_back(structure);
    blist[1] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(1);
    structures.push_back(structure);
    structure.clear();
    structure.push_back(2);
    structure.push_back(3);
    structures.push_back(structure);
    blist[2] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(1);
    structure.push_back(2);
    structure.push_back(3);
    structures.push_back(structure);
    blist[3] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(1);
    structure.push_back(3);
    structure.push_back(2);
    structures.push_back(structure);
    blist[4] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(2);
    structure.push_back(1);
    structure.push_back(3);
    structures.push_back(structure);
    blist[5] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(2);
    structure.push_back(3);
    structure.push_back(1);
    structures.push_back(structure);
    blist[6] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(3);
    structure.push_back(1);
    structure.push_back(2);
    structures.push_back(structure);
    blist[7] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(3);
    structure.push_back(2);
    structure.push_back(1);
    structures.push_back(structure);
    blist[8] = structures;
    return blist;
  }

  if ( basis == id33bar88 ) {
    structures.clear();
    structure.clear();
    structure.push_back(2);
    structure.push_back(3);
    structures.push_back(structure);
    structure.clear();
    structure.push_back(0);
    structure.push_back(1);
    structures.push_back(structure);
    blist[0] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(2);
    structure.push_back(3);
    structure.push_back(1);
    structures.push_back(structure);
    blist[1] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(3);
    structure.push_back(2);
    structure.push_back(1);
    structures.push_back(structure);
    blist[2] = structures;
    return blist;
  }

  if ( basis == id33bar33bar ) {
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(3);
    structures.push_back(structure);
    structure.clear();
    structure.push_back(2);
    structure.push_back(1);
    structures.push_back(structure);
    blist[0] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(1);
    structures.push_back(structure);
    structure.clear();
    structure.push_back(2);
    structure.push_back(3);
    structures.push_back(structure);
    blist[1] = structures;
    return blist;
  }

  if ( basis == id88888 ) {
    structures.clear();
    structure.clear();
    structure.push_back(3);
    structure.push_back(4);
    structures.push_back(structure);
    structure.clear();
    structure.push_back(0);
    structure.push_back(1);
    structure.push_back(2);
    structures.push_back(structure);
    blist[0] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(2);
    structure.push_back(4);
    structures.push_back(structure);
    structure.clear();
    structure.push_back(0);
    structure.push_back(1);
    structure.push_back(3);
    structures.push_back(structure);
    blist[1] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(2);
    structure.push_back(3);
    structures.push_back(structure);
    structure.clear();
    structure.push_back(0);
    structure.push_back(1);
    structure.push_back(4);
    structures.push_back(structure);
    blist[2] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(3);
    structure.push_back(4);
    structures.push_back(structure);
    structure.clear();
    structure.push_back(0);
    structure.push_back(2);
    structure.push_back(1);
    structures.push_back(structure);
    blist[3] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(1);
    structure.push_back(4);
    structures.push_back(structure);
    structure.clear();
    structure.push_back(0);
    structure.push_back(2);
    structure.push_back(3);
    structures.push_back(structure);
    blist[4] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(1);
    structure.push_back(3);
    structures.push_back(structure);
    structure.clear();
    structure.push_back(0);
    structure.push_back(2);
    structure.push_back(4);
    structures.push_back(structure);
    blist[5] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(2);
    structure.push_back(4);
    structures.push_back(structure);
    structure.clear();
    structure.push_back(0);
    structure.push_back(3);
    structure.push_back(1);
    structures.push_back(structure);
    blist[6] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(1);
    structure.push_back(4);
    structures.push_back(structure);
    structure.clear();
    structure.push_back(0);
    structure.push_back(3);
    structure.push_back(2);
    structures.push_back(structure);
    blist[7] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(1);
    structure.push_back(2);
    structures.push_back(structure);
    structure.clear();
    structure.push_back(0);
    structure.push_back(3);
    structure.push_back(4);
    structures.push_back(structure);
    blist[8] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(2);
    structure.push_back(3);
    structures.push_back(structure);
    structure.clear();
    structure.push_back(0);
    structure.push_back(4);
    structure.push_back(1);
    structures.push_back(structure);
    blist[9] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(1);
    structure.push_back(3);
    structures.push_back(structure);
    structure.clear();
    structure.push_back(0);
    structure.push_back(4);
    structure.push_back(2);
    structures.push_back(structure);
    blist[10] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(1);
    structure.push_back(2);
    structures.push_back(structure);
    structure.clear();
    structure.push_back(0);
    structure.push_back(4);
    structure.push_back(3);
    structures.push_back(structure);
    blist[11] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(4);
    structures.push_back(structure);
    structure.clear();
    structure.push_back(1);
    structure.push_back(2);
    structure.push_back(3);
    structures.push_back(structure);
    blist[12] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(3);
    structures.push_back(structure);
    structure.clear();
    structure.push_back(1);
    structure.push_back(2);
    structure.push_back(4);
    structures.push_back(structure);
    blist[13] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(4);
    structures.push_back(structure);
    structure.clear();
    structure.push_back(1);
    structure.push_back(3);
    structure.push_back(2);
    structures.push_back(structure);
    blist[14] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(2);
    structures.push_back(structure);
    structure.clear();
    structure.push_back(1);
    structure.push_back(3);
    structure.push_back(4);
    structures.push_back(structure);
    blist[15] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(3);
    structures.push_back(structure);
    structure.clear();
    structure.push_back(1);
    structure.push_back(4);
    structure.push_back(2);
    structures.push_back(structure);
    blist[16] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(2);
    structures.push_back(structure);
    structure.clear();
    structure.push_back(1);
    structure.push_back(4);
    structure.push_back(3);
    structures.push_back(structure);
    blist[17] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(1);
    structures.push_back(structure);
    structure.clear();
    structure.push_back(2);
    structure.push_back(3);
    structure.push_back(4);
    structures.push_back(structure);
    blist[18] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(1);
    structures.push_back(structure);
    structure.clear();
    structure.push_back(2);
    structure.push_back(4);
    structure.push_back(3);
    structures.push_back(structure);
    blist[19] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(1);
    structure.push_back(2);
    structure.push_back(3);
    structure.push_back(4);
    structures.push_back(structure);
    blist[20] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(1);
    structure.push_back(2);
    structure.push_back(4);
    structure.push_back(3);
    structures.push_back(structure);
    blist[21] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(1);
    structure.push_back(3);
    structure.push_back(2);
    structure.push_back(4);
    structures.push_back(structure);
    blist[22] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(1);
    structure.push_back(3);
    structure.push_back(4);
    structure.push_back(2);
    structures.push_back(structure);
    blist[23] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(1);
    structure.push_back(4);
    structure.push_back(2);
    structure.push_back(3);
    structures.push_back(structure);
    blist[24] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(1);
    structure.push_back(4);
    structure.push_back(3);
    structure.push_back(2);
    structures.push_back(structure);
    blist[25] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(2);
    structure.push_back(1);
    structure.push_back(3);
    structure.push_back(4);
    structures.push_back(structure);
    blist[26] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(2);
    structure.push_back(1);
    structure.push_back(4);
    structure.push_back(3);
    structures.push_back(structure);
    blist[27] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(2);
    structure.push_back(3);
    structure.push_back(1);
    structure.push_back(4);
    structures.push_back(structure);
    blist[28] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(2);
    structure.push_back(3);
    structure.push_back(4);
    structure.push_back(1);
    structures.push_back(structure);
    blist[29] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(2);
    structure.push_back(4);
    structure.push_back(1);
    structure.push_back(3);
    structures.push_back(structure);
    blist[30] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(2);
    structure.push_back(4);
    structure.push_back(3);
    structure.push_back(1);
    structures.push_back(structure);
    blist[31] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(3);
    structure.push_back(1);
    structure.push_back(2);
    structure.push_back(4);
    structures.push_back(structure);
    blist[32] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(3);
    structure.push_back(1);
    structure.push_back(4);
    structure.push_back(2);
    structures.push_back(structure);
    blist[33] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(3);
    structure.push_back(2);
    structure.push_back(1);
    structure.push_back(4);
    structures.push_back(structure);
    blist[34] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(3);
    structure.push_back(2);
    structure.push_back(4);
    structure.push_back(1);
    structures.push_back(structure);
    blist[35] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(3);
    structure.push_back(4);
    structure.push_back(1);
    structure.push_back(2);
    structures.push_back(structure);
    blist[36] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(3);
    structure.push_back(4);
    structure.push_back(2);
    structure.push_back(1);
    structures.push_back(structure);
    blist[37] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(4);
    structure.push_back(1);
    structure.push_back(2);
    structure.push_back(3);
    structures.push_back(structure);
    blist[38] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(4);
    structure.push_back(1);
    structure.push_back(3);
    structure.push_back(2);
    structures.push_back(structure);
    blist[39] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(4);
    structure.push_back(2);
    structure.push_back(1);
    structure.push_back(3);
    structures.push_back(structure);
    blist[40] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(4);
    structure.push_back(2);
    structure.push_back(3);
    structure.push_back(1);
    structures.push_back(structure);
    blist[41] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(4);
    structure.push_back(3);
    structure.push_back(1);
    structure.push_back(2);
    structures.push_back(structure);
    blist[42] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(4);
    structure.push_back(3);
    structure.push_back(2);
    structure.push_back(1);
    structures.push_back(structure);
    blist[43] = structures;
    return blist;
  }

  if ( basis == id33bar888 ) {
    structures.clear();
    structure.clear();
    structure.push_back(2);
    structure.push_back(3);
    structure.push_back(4);
    structures.push_back(structure);
    structure.clear();
    structure.push_back(0);
    structure.push_back(1);
    structures.push_back(structure);
    blist[0] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(2);
    structure.push_back(4);
    structure.push_back(3);
    structures.push_back(structure);
    structure.clear();
    structure.push_back(0);
    structure.push_back(1);
    structures.push_back(structure);
    blist[1] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(3);
    structure.push_back(4);
    structures.push_back(structure);
    structure.clear();
    structure.push_back(0);
    structure.push_back(2);
    structure.push_back(1);
    structures.push_back(structure);
    blist[2] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(2);
    structure.push_back(4);
    structures.push_back(structure);
    structure.clear();
    structure.push_back(0);
    structure.push_back(3);
    structure.push_back(1);
    structures.push_back(structure);
    blist[3] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(2);
    structure.push_back(3);
    structures.push_back(structure);
    structure.clear();
    structure.push_back(0);
    structure.push_back(4);
    structure.push_back(1);
    structures.push_back(structure);
    blist[4] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(2);
    structure.push_back(3);
    structure.push_back(4);
    structure.push_back(1);
    structures.push_back(structure);
    blist[5] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(2);
    structure.push_back(4);
    structure.push_back(3);
    structure.push_back(1);
    structures.push_back(structure);
    blist[6] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(3);
    structure.push_back(2);
    structure.push_back(4);
    structure.push_back(1);
    structures.push_back(structure);
    blist[7] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(3);
    structure.push_back(4);
    structure.push_back(2);
    structure.push_back(1);
    structures.push_back(structure);
    blist[8] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(4);
    structure.push_back(2);
    structure.push_back(3);
    structure.push_back(1);
    structures.push_back(structure);
    blist[9] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(4);
    structure.push_back(3);
    structure.push_back(2);
    structure.push_back(1);
    structures.push_back(structure);
    blist[10] = structures;
    return blist;
  }

  if ( basis == id33bar33bar8 ) {
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(4);
    structure.push_back(3);
    structures.push_back(structure);
    structure.clear();
    structure.push_back(2);
    structure.push_back(1);
    structures.push_back(structure);
    blist[0] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(4);
    structure.push_back(1);
    structures.push_back(structure);
    structure.clear();
    structure.push_back(2);
    structure.push_back(3);
    structures.push_back(structure);
    blist[1] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(3);
    structures.push_back(structure);
    structure.clear();
    structure.push_back(2);
    structure.push_back(4);
    structure.push_back(1);
    structures.push_back(structure);
    blist[2] = structures;
    structures.clear();
    structure.clear();
    structure.push_back(0);
    structure.push_back(1);
    structures.push_back(structure);
    structure.clear();
    structure.push_back(2);
    structure.push_back(4);
    structure.push_back(3);
    structures.push_back(structure);
    blist[3] = structures;
    return blist;
  }

  throw Exception() << "SimpleColourBasis2::basisList(): Cannot handle colour configuration" << Exception::runerror;

  return blist;

}

void SimpleColourBasis2::makeIds() const {

  id88 = vector<PDT::Colour>(2,PDT::Colour8);

  id33bar.push_back(PDT::Colour3);
  id33bar.push_back(PDT::Colour3bar);

  id888 = vector<PDT::Colour>(3,PDT::Colour8);

  id33bar8 = id33bar;
  id33bar8.push_back(PDT::Colour8);

  id8888 = vector<PDT::Colour>(4,PDT::Colour8);

  id33bar88 = id33bar8;
  id33bar88.push_back(PDT::Colour8);

  id33bar33bar = id33bar;
  id33bar33bar.push_back(PDT::Colour3);
  id33bar33bar.push_back(PDT::Colour3bar);

  id88888 = vector<PDT::Colour>(5,PDT::Colour8);

  id33bar888 = id33bar88;
  id33bar888.push_back(PDT::Colour8);

  id33bar33bar8 = id33bar33bar;
  id33bar33bar8.push_back(PDT::Colour8);

}


// If needed, insert default implementations of virtual function defined
// in the InterfacedBase class here (using ThePEG-interfaced-impl in Emacs).

void SimpleColourBasis2::persistentOutput(PersistentOStream &) const {}

void SimpleColourBasis2::persistentInput(PersistentIStream &, int) {}


// *** Attention *** The following static variable is needed for the type
// description system in ThePEG. Please check that the template arguments
// are correct (the class and its base class), and that the constructor
// arguments are correct (the class name and the name of the dynamically
// loadable library where the class implementation can be found).
DescribeClass<SimpleColourBasis2,ColourBasis>
  describeHerwigSimpleColourBasis2("Herwig::SimpleColourBasis2", "Herwig.so");

void SimpleColourBasis2::Init() {

  static ClassDocumentation<SimpleColourBasis2> documentation
    ("SimpleColourBasis2 implements the colour algebra needed for "
     "processes with four coloured legs at NLO. It mainly "
     "serves as an example for the general ColourBasis interface.");

}

